

{-----------------------------------------------------------------------------
 Unit Name: InitParams
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Initializations after entering a formula for a surface
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit InitParams;

interface

procedure ComputeVPCubeKey;
procedure ComputeFocusCubeKey;
procedure InitializeSurfaceParameters(comesFromFile: Boolean);
procedure SetFocus;
procedure FrameFromFocus;
procedure ComputeCube;

implementation
uses
   Math,
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}
{$IFDEF LINUX}
  Types,
{$ENDIF}

Globals, FSurface, OCFunct, Utilities, ParallelFrame, ManageClips;

procedure ComputeVPCubeKey;
var
  i: Integer;
begin
  for i := 1 to 3 do begin
    gVPCubeKey[i] := 2;
    if gViewpoint[i] > gMax3D[i] then
      gVPCubeKey[i] := 1;
    if gViewpoint[i] < gMin3D[i] then
      gVPCubeKey[i] := 0;
  end;
end;

procedure ComputeFocusCubeKey;
var
  i: Integer;
begin
  for i := 1 to 3 do begin
    gFCubeKey[i] := 2;
    if gFocus[i] > gMax3D[i] then
      gFCubeKey[i] := 1;
    if gFocus[i] < gMin3D[i] then
      gFCubeKey[i] := 0;
  end;
end;

procedure Initialize3DMetrics;
var
  k: Integer;
begin
  for k := 1 to 3 do begin
    gEdges3D[k] := gMax3D[k] - gMin3D[k];
    gHalfEdges3D[k] := gEdges3D[k] / 2;
    gCenter3D[k] := (gMax3D[k] + gMin3D[k]) / 2;
    gIncr3D[k] := gEdges3D[k] / 32000;
  end;
  gSpan3D := EuclNorm3(gEdges3D);
  gRadius3D := gSpan3D / (kWidth + kHeight);
  gMaxSegmentLength3D := 8 * gRadius3D;
  gTolDistance := 2 * sqr(gSpan3D / kPPLine2D);
  gVerySmall := sqr(gRadius3D) / 8;
end;

procedure ComputeMinMax3D;
var
  i, j, k: Integer;
  param, step: Vector2;
  p: vector3;
begin
  for i := 1 to 3 do begin
    gMin3D[i] := 1.0E20;
    gMax3D[i] := -1.0E20;
  end;
  step[1] := gEdges2D[1] / 20;
  step[2] := gEdges2D[2] / 20;
  for i := 0 to 20 do
    for j := 0 to 20 do begin
      param[1] := gMin2D[1] + i * step[1];
      param[2] := gMin2D[2] + j * step[2];
      theFPar2(p, param);
      for k := 1 to 3 do begin
        if p[k] < gMin3D[k] then
          gMin3D[k] := p[k];
        if p[k] > gMax3D[k] then
          gMax3D[k] := p[k];
      end;
    end;
  Initialize3DMetrics;
end;

procedure SetViewpoint;
const
  farthness: double = 3.5;
var
  distance, alpha: double;
begin
  distance := farthness * gSpan3D;
  if gSpan2D < 1.0E-9 then
    alpha := 0
  else
    alpha := arcsin(gEdges3D[2] / gSpan3D);
  alpha := 0.5 * (alpha + pi / 4);
  if (abs(alpha - pi / 4) < 15) then
    alpha := pi / 4 + 0.15;
  gViewpoint[1] := gMin3D[1] + cos(alpha) * distance;
  gViewpoint[2] := gMin3D[2] + sin(alpha) * distance;
  gViewpoint[3] := gMax3D[3] + 0.4 * distance;
end;

procedure SetFocus;
var
  r: Double;
  i: Integer;
begin
  r := 0.3 * EuclDistance3(gCenter3D, gViewpoint);
  for i := 1 to 3 do
    gFocus[i] := gViewpoint[i] - r * (gH[i] + gV[i]);
end;

procedure FrameFromViewpoint;
var
  i, j, k: Integer;
  maximum, minimum: Vector2;
  q: double;
  pos: Vector3;

  procedure Sieve(u: Vector3);
  var
    z: Vector2;
    ii: Integer;
  begin
    z := ProjectToPaneReal(u);
    for ii := 1 to 2 do begin
      if z[ii] > maximum[ii] then
        maximum[ii] := z[ii];
      if z[ii] < minimum[ii] then
        minimum[ii] := z[ii]
    end;
  end; {Sieve}

begin {FrameFromViewpoint}
  q := EuclDistance3(gCenter3D, gViewPoint);
  for i := 1 to 3 do
    gN[i] := (gCenter3D[i] - gViewpoint[i]) / q;
  q := hypot(gN[1], gN[2]);
  if (q > 1.0E-6) then begin
    gH[1] := gN[2] / q;
    gH[2] := -gN[1] / q;
    gH[3] := 0.0;
    gV[1] := -gH[2] * gN[3];
    gV[2] := gH[1] * gN[3];
    gV[3] := -q;
  end
  else begin
    gH[1] := 1.0;
    gH[2] := 0.0;
    gH[3] := 0.0;
    gV[1] := 0.0;
    gV[2] := 1.0;
    gV[3] := 0.0;
  end;
  gC0 := 1;
  gC1 := EuclInnerProd3(gViewpoint, gH);
  gC2 := EuclInnerProd3(gViewpoint, gV);
  gC3 := EuclInnerProd3(gViewpoint, gN);
  gpcx := kHalfWidth;
  gpcy := kHalfHeight;

  maximum[1] := -maxInt;
  maximum[2] := -maxInt;
  minimum[1] := maxInt;
  minimum[2] := maxInt;
  for i := 0 to 1 do
    for j := 0 to 1 do
      for k := 0 to 1 do begin
        pos[1] := gMin3D[1] + i * gEdges3D[1];
        pos[2] := gMin3D[2] + j * gEdges3D[2];
        pos[3] := gMin3D[3] + k * gEdges3D[3];
        Sieve(pos);
      end;
  gC0 := 0.95 * min(kWidth / (maximum[1] - minimum[1]), kHeight /
    (maximum[2] - minimum[2]));
end;

procedure FrameFromFocus;
var
  i, j, k: Integer;
  maximum, minimum: Vector2;
  q: double;
  pos: Vector3;

  procedure Sieve(u: Vector3);
  var
    z: Vector2;
    ii: Integer;
  begin
    z := ProjectToFocusPlaneReal(u);
    for ii := 1 to 2 do begin
      if z[ii] > maximum[ii] then
        maximum[ii] := z[ii];
      if z[ii] < minimum[ii] then
        minimum[ii] := z[ii]
    end;
  end; {Sieve}

begin {FrameFromFocus}
  q := EuclDistance3(gCenter3D, gFocus);
  for i := 1 to 3 do
    gFN[i] := (gCenter3D[i] - gFocus[i]) / q;
  q := hypot(gFN[1], gFN[2]);
  if (q > 1.0E-6) then begin
    gFH[1] := gFN[2] / q;
    gFH[2] := -gFN[1] / q;
    gFH[3] := 0.0;
    gFV[1] := -gFH[2] * gFN[3];
    gFV[2] := gFH[1] * gFN[3];
    gFV[3] := -q;
  end
  else begin
    gFH[1] := 1.0;
    gFH[2] := 0.0;
    gFH[3] := 0.0;
    gFV[1] := 0.0;
    gFV[2] := 1.0;
    gFV[3] := 0.0;
  end;
  gFC0 := 1;
  gFC1 := EuclInnerProd3(gFocus, gFH);
  gFC2 := EuclInnerProd3(gFocus, gFV);
  gFC3 := EuclInnerProd3(gFocus, gFN);

  maximum[1] := -maxInt;
  maximum[2] := -maxInt;
  minimum[1] := maxInt;
  minimum[2] := maxInt;
  for i := 0 to 1 do
    for j := 0 to 1 do
      for k := 0 to 1 do begin
        pos[1] := gMin3D[1] + i * gEdges3D[1];
        pos[2] := gMin3D[2] + j * gEdges3D[2];
        pos[3] := gMin3D[3] + k * gEdges3D[3];
        Sieve(pos);
      end;
  gFC0 := 0.95 * min(kWidth / (maximum[1] - minimum[1]), kHeight /
    (maximum[2] - minimum[2]));
end;

procedure AddSegmentCube(x1, x2: Vector3);
var
  s1, s2: Vector2;
begin
  inc(gCubeLast);
  if High(gCube) < gCubeLast then
    SetLength(gCube, gCubeLast + 10);
  s1 := ProjectToPaneReal(x1);
  s2 := ProjectToPaneReal(x2);
  gCube[gCubeLast, 1] := s1;
  gCube[gCubeLast, 2] := s2;
end;

function IsVisible(y: Vector3): Boolean;
var
  pixel: TPoint;
  d: Single;
begin
  pixel := ProjectToPaneIntAndDist(y, d);
  if (gType = Parametric) or (gType = Tube) then
    result := PtInRect(gRect, pixel) and (d < 1.001 *
      gZBuffer^[pixel.x, pixel.y])
  else
    result := PtInRect(gRect, pixel) and IsVisibleImplicit(gViewPoint, y,
      gVPCubeKey, False);
end;


procedure ExamineLine(x: Vector3; c: Integer);
const
  cStepNumber = 100;
var
  step: Double;
  y, s1, s2: Vector3;
  i: Integer;
  foundFirst, visible: Boolean;
begin
  step := gEdges3D[c] / cStepNumber;
  y := x;
  foundFirst := IsVisible(y);
  if foundFirst then
    s1 := y;
  s2 := y;
  for i := 1 to cStepNumber do begin
    y[c] := x[c] + i * step;
    visible := IsVisible(y);
    if visible and not foundFirst then begin
      foundFirst := true;
      s1 := y;
    end
    else if foundFirst and (not visible) then begin
      foundFirst := False;
      AddSegmentCube(s1, s2);
    end;
    s2 := y;
  end;
  if FoundFirst and Visible then
    AddSegmentCube(s1, s2);
end;

procedure ComputeCube;
var
  first, second, third, i, j: Integer;
  x1, x2: Vector3;
begin
  gCube := nil;
  gCubeLast := -1;
  for first := 1 to 3 do begin
    second := (first mod 3) + 1;
    third := (second mod 3) + 1;
    for i := 0 to 1 do
      for j := 0 to 1 do begin
        x1[first] := gMin3D[first];
        x1[second] := gMin3D[second] + i * gEdges3D[second];
        x1[third] := gMin3D[third] + j * gEdges3D[third];
        x2[first] := gMax3D[first];
        x2[second] := x1[second];
        x2[third] := x1[third];
        if (gVPCubeKey[second] = i) or (gVPCubeKey[third] = j) then
          AddSegmentCube(x1, x2)
        else
          ExamineLine(x1, first);
      end;
  end;
end;


procedure FrameParamSurface;
begin
  ComputeMinMax3D;
  if not gGotProjParams then begin
    SetViewpoint;
    FrameFromViewpoint;
  end;
  ComputeVPCubeKey;
end;

procedure InitParametricSurfParams;
var
  i: Integer;
begin
  theFPar := TheParamMap;
  theFPar2 := TheParamMap2;

  for i := 1 to 2 do begin
    gCenter2D[i] := (gMax2D[i] + gMin2D[i]) / 2;
    gEdges2D[i] := gMax2D[i] - gMin2D[i];
    gHalfEdges2D[i] := gEdges2D[i] / 2;
    gIncr2D[i] := gEdges2D[i] / 80000;
    g4Incr2D[i] := 4 * gIncr2D[i];
    g2SqrIncr2D[i] := 2 * sqr(gIncr2D[i]);
  end;
  gStep2D[1] := gEdges2D[1] / gGridU;
  gStep2D[2] := gEdges2D[2] / gGridV;
  g4Incr1Incr2 := 4 * gIncr2D[1] * gIncr2D[2];
  gSpan2D := hypot(gEdges2D[1], gEdges2D[2]);
  gRadius2D := gSpan2D / (kWidth + kHeight);
  gMaxSegmentLength2D := 7 * gRadius2D;

  pd1 := kWidth / gEdges2D[1];
  pd2 := -pd1 * gMin2D[1];
  pd3 := -kHeight / gEdges2D[2];
  pd4 := -pd3 * gMax2D[2];

  dp1 := 1 / pd1;
  dp2 := -pd2 * dp1;
  dp3 := 1 / pd3;
  dp4 := -pd4 * dp3;
  FrameParamSurface;

  gDist := EuclDistance3(gViewpoint, gCenter3D);
end;

procedure EstimateBestParametersImplicitSurf;
const
  samplesPerLine = 7;
  maxDivisorEpsilon: double = 100000.0;
  minDivisorEpsilon: double = 15000.0;
  littleReal: double = 1.0E-12;
  minPPLine2D = 15;
  minnumFlags = 201;

var
  epsi, place, diagonal: Vector3;
  meanVariation, divisorEpsilon, average1, z, ruggedness1, toAdd, absSum:
  double;
  i, j, k, c, d: Integer;
  values: array[1..samplesPerLine, 1..SamplesPerLine, 1..SamplesPerLine] of
  double;
begin
  average1 := 0;
  for i := 1 to 3 do begin
    diagonal[i] := gMax3D[i] - gMin3D[i];
    epsi[i] := diagonal[i] / (2 * samplesPerLine);
    gIncr3D[i] := diagonal[i] / maxDivisorEpsilon;
  end;

  ruggedness1 := 0;
  for i := 1 to samplesPerLine do
    for j := 1 to samplesPerLine do
      for k := 1 to samplesPerLine do begin
        place[1] := gMin3D[1] + (2 * i - 1) * epsi[1];
        place[2] := gMin3D[2] + (2 * j - 1) * epsi[2];
        place[3] := gMin3D[3] + (2 * k - 1) * epsi[3];
        values[i, j, k] := gTheF1.V(place);
        average1 := average1 + abs(values[i, j, k]);
      end;
  for i := 2 to samplesPerLine - 1 do
    for j := 2 to samplesPerLine - 1 do
      for k := 2 to samplesPerLine - 1 do begin
        z := 2 * values[i, j, k];
        abssum := abs(values[i, j, k]) + abs(values[i - 1, j, k]) +
          abs(values[i + 1, j, k]) + abs(values[i, j - 1, k]) + abs(values[i, j
          + 1, k]) + abs(values[i, j, k - 1]) + abs(values[i, j, k + 1]);
        toAdd := abs(values[i + 1, j, k] + values[i - 1, j, k] - z) +
          abs(values[i, j + 1, k] + values[i, j - 1, k] - z) + abs(values[i,
          j, k
            + 1] + values[i, j, k - 1] - z);
        ruggedness1 := ruggedness1 + toAdd / (absSum + littleReal);
      end;
  c := samplesPerLine * sqr(samplesPerLine);
  d := (samplesPerLine - 2) * sqr(samplesPerLine - 2);
  meanVariation := ruggedness1 / d;
  meanVariation := meanVariation / (sqr(epsi[1]) + sqr(epsi[2]) +
    sqr(epsi[3]));
  average1 := average1 / c;
  gVerySmall := min(average1 / 100000, 1.0E-7);
  gVerySmall := max(gVerySmall, 1.0E-9);

  divisorEpsilon := 3.0 * minDivisorEpsilon * meanVariation;
  if divisorEpsilon < minDivisorEpsilon then
    divisorEpsilon := minDivisorEpsilon;
  if divisorEpsilon > maxDivisorEpsilon then
    divisorEpsilon := maxDivisorEpsilon;
  for i := 1 to 3 do
    gIncr3D[i] := diagonal[i] / divisorEpsilon;
  gPPLine2D := round(3 * meanVariation);
  if gPPLine2D > kmaxPPLine2D then
    gPPLine2D := kmaxPPLine2D;
  if gPPLine2D < minPPLine2D then
    gPPLine2D := minPPLine2D;
  if not odd(gPPLine2D) then
    gPPLine2D := gPPLine2D - 1;

  gNumFlags := 4 * gPPLine2D;
  if gNumFlags > kmaxnumFlags then
    gNumFlags := kmaxnumFlags;
  if gNumFlags < minnumFlags then
    gNumFlags := minnumFlags;
  if not odd(gNumFlags) then
    gNumFlags := gNumFlags - 1;
end;

procedure InitImplicitSurfParams;
begin
  Initialize3DMetrics;
  if not gGotProjParams then begin
    SetViewpoint;
    FrameFromViewpoint;
  end;
  EstimateBestParametersImplicitSurf;
  ComputeVPCubeKey;
  gDist := EuclDistance3(gViewpoint, gCenter3D);
end;

procedure InitTubeParamsAndFunctions;
var
  i: Integer;
begin
  for i := 1 to 2 do begin
    gEdges2D[i] := gMax2D[i] - gMin2D[i];
    gCenter2D[i] := (gMax2D[i] + gMin2D[i]) / 2;
    gHalfEdges2D[i] := gEdges2D[i] / 2;
    gIncr2D[i] := gEdges2D[i] / 80000;
    g2SqrIncr2D[i] := 2 * sqr(gIncr2D[i]);
  end;
  gStep2D[2] := gEdges2D[2] / gGridV;
  BuildParallelFrame;
  theFPar := TheTube;
  theFPar2 := TheTube2;
  for i := 1 to 2 do begin
    g4Incr2D[i] := 4 * gIncr2D[i];
    g2SqrIncr2D[i] := 2 * sqr(gIncr2D[i]);
  end;
  g4Incr1Incr2 := 4 * gIncr2D[1] * gIncr2D[2];
  gSpan2D := hypot(gEdges2D[1], gEdges2D[2]);
  gRadius2D := gSpan2D / (kWidth + kHeight);
  gMaxSegmentLength2D := 7 * gRadius2D;

  pd1 := kWidth / gEdges2D[1];
  pd2 := -pd1 * gMin2D[1];
  pd3 := -kHeight / gEdges2D[2];
  pd4 := -pd3 * gMax2D[2];

  dp1 := 1 / pd1;
  dp2 := -pd2 * dp1;
  dp3 := 1 / pd3;
  dp4 := -pd4 * dp3;

  FrameParamSurface;
  gDist := EuclDistance3(gViewpoint, gCenter3D);
end;

procedure EstimateBestParametersPseudoRiemannianSurf;
const
  samplesPerLine = 7;
  maxDivisorEpsilon: double = 50000.0;
  minDivisorEpsilon: double = 5000.0;
  littleReal: double = 1.0E-12;
var
  epsi, place: Vector2;
  firstFF: Vector3;
  meanVariation, divisorEpsilon, average1, z, ruggedness1, toAdd, absSum:
  double;
  i, j, c, d: Integer;
  values: array[1..samplesPerLine, 1..SamplesPerLine] of double;
begin
  for i := 1 to 2 do begin
    gEdges2D[i] := gMax2D[i] - gMin2D[i];
    gHalfEdges2D[i] := gEdges2D[i] / 2;
    gCenter2D[i] := (gMax2D[i] + gMin2D[i]) / 2;
    epsi[i] := gEdges2D[i] / (2 * samplesPerLine);
    gIncr2D[i] := gEdges2D[i] / maxDivisorEpsilon;
  end;
  gSpan2D := hypot(gEdges2D[1], gEdges2D[2]);
  average1 := 0;
  ruggedness1 := 0;
  for i := 1 to samplesPerLine do
    for j := 1 to samplesPerLine do begin
      place[1] := gMin2D[1] + (2 * i - 1) * epsi[1];
      place[2] := gMin2D[2] + (2 * j - 1) * epsi[2];
      firstFF[1] := gTheF1.V(place);
      firstFF[2] := gTheF2.V(place);
      firstFF[3] := gTheF3.V(place);
      values[i, j] := firstFF[1] + firstFF[2] + firstFF[3];
      average1 := average1 + abs(values[i, j]);
    end;
  for i := 2 to samplesPerLine - 1 do
    for j := 2 to samplesPerLine - 1 do begin
      z := 2 * values[i, j];
      abssum := values[i, j] + values[i - 1, j] + values[i + 1, j] + values[i,
        j - 1] + values[i, j + 1];
      toAdd := abs(values[i + 1, j] + values[i - 1, j] - z) + abs(values[i, j
        + 1] + values[i, j - 1] - z);
      ruggedness1 := ruggedness1 + toAdd / (absSum + littleReal);
    end;
  c := samplesPerLine * sqr(samplesPerLine);
  d := (samplesPerLine - 2) * sqr(samplesPerLine - 2);
  meanVariation := ruggedness1 / d;
  meanVariation := meanVariation / (sqr(epsi[1]) + sqr(epsi[2]));
  average1 := average1 / c;
  gVerySmall := min(average1 / 100000, 1.0E-7);
  gVerySmall := max(gVerySmall, 1.0E-9);

  divisorEpsilon := 3.0 * minDivisorEpsilon * meanVariation;
  if divisorEpsilon < minDivisorEpsilon then
    divisorEpsilon := minDivisorEpsilon;
  if divisorEpsilon > maxDivisorEpsilon then
    divisorEpsilon := maxDivisorEpsilon;
  for i := 1 to 2 do begin
    gIncr2D[i] := gEdges2D[i] / divisorEpsilon;
    g2Incr2D[i] := 2 * gIncr2D[i];
    g4Incr2D[i] := 4 * gIncr2D[i];
    g24Incr2D[i] := 24 * gIncr2D[i];
    g24SqrIncr2D[i] := g24Incr2D[i] * gIncr2D[i];
  end;
  g4Incr1Incr2 := 4 * gIncr2D[1] * gIncr2D[2];
  gRadius2D := gSpan2D / (kWidth + kHeight);
  gMaxSegmentLength2D := 5 * gRadius2D;

  pd1 := kWidthMinus1 / gEdges2D[1];
  pd3 := -kHeightMinus1 / gEdges2D[2];
  if pd1 < abs(pd3) then begin
    pd2 := -pd1 * gMin2D[1];
    pd3 := -pd1;
    pd4 := kHalfHeight - pd3 * gCenter2D[2]
  end
  else if pd1 > abs(pd3) then begin
    pd1 := -pd3;
    pd4 := -pd3 * gMax2D[2];
    pd2 := kHalfWidth - pd1 * gCenter2D[1];
  end
  else begin
    pd2 := -pd1 * gMin2D[1];
    pd4 := -pd3 * gMax2D[2];
  end;
  dp1 := 1 / pd1;
  dp2 := -pd2 / pd1;
  dp3 := 1 / pd3;
  dp4 := -pd4 / pd3;
end;

procedure InitPseudoRiemannianParams;
begin
  EstimateBestParametersPseudoRiemannianSurf;
end;

procedure InitializeSurfaceParameters;
begin
  case gType of
    Parametric:
      InitParametricSurfParams;
    Implicit:
      InitImplicitSurfParams;
    PseudoRiemannian:
      InitPseudoRiemannianParams;
    Tube:
      InitTubeParamsAndFunctions;
  else
    ;
  end;
  if (gType <> PseudoRiemannian) then begin
    PrepareToSetClip;
    if (not gGotProjParams) then begin
      gClips := nil;
      gClipTop := -1;
    end;
  end;
end;
end.
