

{-----------------------------------------------------------------------------
 Unit Name: ParallelFrame
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: For drawing a tube around a curve, I first compute an orthonormal
    frame parallel along the curve, one of whose vectors is tangent to it.
    The other two vectors will serve for parameterizing the circles that
    make the tube.
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}



unit ParallelFrame;

interface
uses
  Globals;

procedure BuildParallelFrame;


procedure TheTube(var tubePt: Vector3; u, v: double);
procedure TheTube2(var tubePt: Vector3; here: vector2);

implementation

uses
  OcFunct, Utilities,
  Math;
const
  uDir: Vector3 = (0.7654300213, -0.3567223309, 0.535598787461241);
  MaxNumberOfPoints = 600;
var
  uRNP: array[0..MaxNumberOfPoints] of Segment3;
  uNumPNF: Integer;
  jump, twoJump, twoJump2: double;
  vel, accel: Vector3;

procedure FirstDerivative(theNormal: Vector3; t: double; var tubePt:
  Vector3);
var
  i: Integer;
  x, xPlus, xMinus: Vector3;
  aux: double;
begin
  TheCurve(x, t);
  TheCurve(xPlus, t + gIncr2D[1]);
  TheCurve(xMinus, t - gIncr2D[1]);
  for i := 1 to 3 do begin
    vel[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[1]);
    accel[i] := 2 * (xPlus[i] + xMinus[i] - 2 * x[i]) / g2SqrIncr2D[1];
  end;
  aux := EuclInnerProd3(vel, vel);
  if aux < 1.0E-12 then
    for i := 1 to 3 do
      tubePt[i] := 0
  else begin
    aux := EuclInnerProd3(theNormal, accel) / aux;
    for i := 1 to 3 do
      tubePt[i] := -aux * vel[i];
  end;
end;

function ParallelNormalRunKut(var t, step, estimStep: double;
  var y, rk1, rk2, rk3: Vector3): boolean;
const
  tol = 2.0E-5;
var
  err, coef: double;
  newY: Vector3;
  keepGoing: boolean;
  counter: integer;

  function CalcElem(t, guesStep: double;
    var z: Vector3): double;
  var
    i: Integer;
    u: Vector3;
  begin
    u := z;
    FirstDerivative(u, t, rk1);
    for i := 1 to 3 do
      u[i] := z[i] + guesStep * rk1[i];
    FirstDerivative(u, t + guesStep, rk2);
    for i := 1 to 3 do
      u[i] := z[i] + guesStep * (rk1[i] + rk2[i]) / 4;
    FirstDerivative(u, t + guesStep / 2, rk3);
    Result := 1.0E-15;
    for i := 1 to 3 do begin
      z[i] := z[i] + guesStep * (rk1[i] + rk2[i] + 4.0 * rk3[i]) / 6.0;
      Result := Result + abs(2 * rk3[i] - rk1[i] - rk2[i]);
    end;
  end; {of CalcElem}

begin {ParallelNormalRunKut}
  counter := 0;
  repeat
    newY := y;
    counter := counter + 1;
    err := CalcElem(t, estimStep, newY);
    coef := sqrt(tol / err);
    if coef < 0.5 then
      coef := 0.5;
    if 2.0 < coef then
      coef := 2.0;
    step := estimStep;
    estimStep := coef * step;
    keepGoing := (counter < 20);
  until (err < tol) or (not keepGoing);
  if keepGoing then begin
    y := newY;
    t := t + step;
  end; {of if keepGoing}
  Result := keepGoing and (abs(step) > 1.0E-12);
end; {de ParallelNormalRunKut}

procedure InterpolateParallelNormal(y, rk1, rk2, rk3: Vector3;
  t, h, theta: double;
  var normals: Segment3);

var
  theta2, b1, b2, b3: double;
  i: Integer;
  xPlus, xMinus, v: Vector3;
begin
  TheCurve(xPlus, t + gIncr2D[1]);
  TheCurve(xMinus, t - gIncr2D[1]);
  theta2 := sqr(theta);
  b1 := theta + (-kThreeOver2 + kTwoOver3 * theta) * theta2;
  b2 := (-0.5 + kTwoOver3 * theta) * theta2;
  b3 := (2.0 - kFourOver3 * theta) * theta2;
  for i := 1 to 3 do begin
    normals[1][i] := y[i] + h * (b1 * rk1[i] + b2 * rk2[i] + b3
      * rk3[i]);
    v[i] := (xPlus[i] - xMinus[i]) / gIncr2D[1];
  end;
  normals[2] := EuclCrossProduct(v, normals[1]);
  b1 := sqrt(EuclInnerProd3(normals[2], normals[2]));
  for i := 1 to 3 do
    normals[2][i] := normals[2][i] / b1;
end;




procedure BuildParallelFrame;
var
  worked: boolean;
  y, prevY, rk1, rk2, rk3, xPlus, xMinus: Vector3;
  t, step, estimStep, theta, prevT,
    tDiscrProx, firstEffectiveT, aux, totalCurv: double;
  i: Integer;

begin
  uNumPNF := 0;
  firstEffectiveT := gMin2D[1] - gIncr2D[1];
  repeat
    firstEffectiveT := firstEffectiveT + gIncr2D[1];
    TheCurve(xPlus, firstEffectiveT + gIncr2D[1]);
    TheCurve(xMinus, firstEffectiveT - gIncr2D[1]);
    for i := 1 to 3 do
      rk1[i] := (xPlus[i] - xMinus[i]) / gIncr2D[1];
    y := EuclCrossProduct(uDir, rk1);
    theta := Sqrt(EuclInnerProd3(y, y));
    worked := firstEffectiveT < gMax2D[1];
  until (theta > 1.0E-7) or not worked;
  if not worked then
    gEdges2D[1] := 0
  else begin
    t := firstEffectiveT;
    for i := 1 to 3 do
      y[i] := y[i] / theta;
    uRNP[uNumPNF][1] := y;
    rk1 := EuclCrossProduct(rk1, y);
    theta := sqrt(EuclInnerProd3(rk1, rk1));
    for i := 1 to 3 do
      rk1[i] := rk1[i] / theta;
    uRNP[uNumPNF][2] := rk1;
    jump := (gMax2D[1] - firstEffectiveT) / MaxNumberOfPoints;
    twoJump := 2 * jump;
    twoJump2 := 2 * sqr(jump);
    inc(uNumPNF);
    tDiscrProx := firstEffectiveT + uNumPNF * jump;
    step := jump;
    estimStep := step;
    totalCurv := 0;
    repeat
      prevT := t;
      prevY := y;
      worked := ParallelNormalRunKut(t, step, estimStep, y, rk1,
        rk2, rk3);
      aux := sqrt(sqr(vel[2] * accel[3] - vel[3] * accel[2]) +
        sqr(vel[3] * accel[1] - vel[1] * accel[3]) +
        sqr(vel[1] * accel[2] - vel[2] * accel[1]));
      aux := aux * step / (sqr(vel[1]) + sqr(vel[2]) + sqr(vel[3]));
      totalCurv := totalCurv + aux;
      if worked then
        while (tDiscrProx <= t) and (uNumPNF <= MaxNumberOfPoints) do begin
          theta := (tDiscrProx - prevT) / step;
          InterpolateParallelNormal(prevY, rk1, rk2, rk3, tDiscrProx,
            step,
            theta, uRNP[uNumPNF]);
          inc(uNumPNF);
          tDiscrProx := gMin2D[1] + uNumPNF * jump;
        end; {de ciclo i}
    until (not worked) or (uNumPNF > MaxNumberOfPoints);
    tDiscrProx := tDiscrProx - jump;
    if gMax2D[1] > tDiscrProx then
      gMax2D[1] := tDiscrProx;
    if gMin2D[1] < firstEffectiveT then
      gMin2D[1] := firstEffectiveT;
    gEdges2D[1] := gMax2D[1] - gMin2D[1];
    totalCurv := totalCurv / (4 * pi);
    if totalCurv > 1 then
      gGridU := round(kTubeGridU * sqrt(totalCurv));
    gStep2D[1] := (gMax2D[1] - gMin2D[1]) / gGridU;
    gIncr2D[1] := (gMax2D[1] - gMin2D[1]) / 80000;
    gHalfEdges2D[1] := gEdges2D[1] / 2;
    gCenter2D[1] := (gMax2D[1] + gMin2D[1]) / 2;
  end;
end; {of BuildParallelFrame}

procedure InterpolateRef(ind: integer; shift: double; var nor, bin: Vector3);
var
  a, b: Vector3;
  i: integer;
  sh2: double;
begin
  sh2 := sqr(shift);
  for i := 1 to 3 do begin
    a[i] := (uRNP[ind + 1][1][i] + uRNP[ind - 1][1][i] - 2 * uRNP[ind][1][i])
      / twoJump2;
    b[i] := (uRNP[ind + 1][1][i] - uRNP[ind - 1][1][i]) / twoJump;
    nor[i] := a[i] * sh2 + b[i] * shift + uRNP[ind][1][i];
    a[i] := (uRNP[ind + 1][2][i] + uRNP[ind - 1][2][i] - 2 * uRNP[ind][2][i])
      / twoJump2;
    b[i] := (uRNP[ind + 1][2][i] - uRNP[ind - 1][2][i]) / twoJump;
    bin[i] := a[i] * sh2 + b[i] * shift + uRNP[ind][2][i];
  end;
end;

procedure TheTube(var tubePt: Vector3; u, v: double);
var
  x, nor, bin: Vector3;
  shift, r: double;
  i, ind: Integer;
  sinus, cosinus: Extended;
begin
  TheCurve(x, u);
  ind := trunc((u - gMin2D[1]) / jump);
  if ind < 1 then
    ind := 1;
  if ind >= MaxNumberOfPoints then
    ind := MaxNumberOfPoints - 1;
  shift := u - gMin2D[1] - jump * ind;
  InterpolateRef(ind, shift, nor, bin);
  r := gTheF4.V([u]);
  sincos(v, sinus, cosinus);
  for i := 1 to 3 do
    tubePt[i] := x[i] + r * (cosinus * nor[i] + sinus * bin[i]);
end;

procedure TheTube2(var tubePt: Vector3; here: vector2);
var
  x, nor, bin: Vector3;
  shift, r: double;
  cosinus, sinus: Extended;
  i, ind: Integer;
begin
  TheCurve(x, here[1]);
  ind := trunc((here[1] - gMin2D[1]) / jump);
  if ind < 1 then
    ind := 1;
  if ind >= MaxNumberOfPoints then
    ind := MaxNumberOfPoints - 1;
  shift := here[1] - gMin2D[1] - jump * ind;
  InterpolateRef(ind, shift, nor, bin);
  r := gTheF4.V([here[1]]);
  sincos(here[2], sinus, cosinus);
  for i := 1 to 3 do
    tubePt[i] := x[i] + r * (cosinus * nor[i] + sinus * bin[i]);
end;

end.
