

{-----------------------------------------------------------------------------
 Unit Name: SecondOrderCurves
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Routines that call de ODE solver for drawing geodesics, etc.
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit SecondOrderCurves;

interface
uses
  Globals;

procedure DrawSecondOrderCurve(toRight: Integer);

implementation

uses
 Math,
{$IFDEF MSWINDOWS}
  Windows, Forms,
{$ENDIF}
{$IFDEF LINUX}
  Types, QForms,
{$ENDIF}
  Utilities, OCFunct, ImplicitUtils, ParametricUtils, BasicRiemUtils, FRunKut;
const
  maxAmpl: double = 0.15;

procedure DrawGeodesicQuadrant(x, e1, e2: Vector3; topT0, beta, gamma:
  Double);
const
  ep: double = 1.0E-3;
var
  alpha, angle, amplitude, guessAmpl, separation, halfAngle, maxAngle, E, F,
    G: double;
  si, co: Extended;
  i, sign: Integer;
  p: vector2;
  x1, x2, vel: Vector3;
  key, prevkey: Integer;
  recDr: RecordAndDrawProcedure;
  good, prevGood: Boolean;

  procedure SetVelPos;
  var
    a: Double;
    k: Integer;
  begin
    sincos(angle, si, co);
    for k := 1 to 3 do
      vel[k] := e1[k] * co + e2[k] * si;
    if gType <> Implicit then
      a := topT0 / sqrt(abs(E * sqr(vel[1]) + 2 * F * vel[1] * vel[2] + G *
        sqr(vel[2])))
    else
      a := topT0 / sqrt(abs(InnerProd3(vel, vel)));
    for k := 1 to 3 do
      vel[k] := a * vel[k];
    x2 := x;
  end;

begin
  case gType of
    Parametric, Tube: begin
        recDr := RecordAndDrawSegmentParamSurface23D;
        theRunKut.SetEnvironment(2, 2, kTol, GeodAccelParametricSurf,
          CheckRangeParametric,
          nil, CheckCameBackParametricSurface, nil);
      end;
    PseudoRiemannian: begin
        recDr := RecordAndDrawSegmentPseudoRiemPlane;
        theRunKut.SetEnvironment(2, 2, kTol, GeodAccelPseudoRiemPlane,
          CheckRangePseudoRiem,
          nil, CameBackPseudoRiemPlane, nil);
      end;
    Implicit: begin
        recDr := RecordAndDrawSegmentImplicitSurface;
        theRunKut.SetEnvironment(3, 2, kTol, GeodAccelImplicitSurf,
          CheckRangeImplicit,
          nil, CheckCameBackImplicitSurface, AdjustPointImplicitSurf2);
      end;
  end;

  if (gType = Parametric) or (gType = Tube) then begin
    p[1] := x[1]; p[2] := x[2];
    ComputeEFG(p, E, F, G);
  end
  else if gType = PseudoRiemannian then begin
    p[1] := x[1]; p[2] := x[2];
    E := gTheF1.V(p);
    F := gTheF2.V(p);
    G := gTheF3.V(p);
  end;
  halfAngle := (beta + gamma) / 2;
  maxAngle := abs(gamma - beta) / 2;
  for i := 0 to 1 do begin
    angle := halfAngle;
    alpha := halfAngle;
    sign := 1 - 2 * i;
    SetVelPos;
    prevKey := theRunKut.ComputeIntegralCurve(x2, vel, 1);
    guessAmpl := maxAmpl;
    x1 := x2;
    repeat
      repeat
        angle := alpha + sign * guessAmpl;
        SetVelPos;
        key := theRunKut.ComputeIntegralCurve(x2, vel, 1);
        amplitude := guessAmpl;
        good := (key and kTopNotReached) = 0;
        prevGood := (prevKey and kTopNotReached) = 0;
        if good or prevGood then begin
          if gType = Implicit then
            separation := 1 + EuclDistance3(x2, x1) / gRadius3D
          else
            separation := 1 + hypot(x2[1] - x1[1], x2[2] - x1[2]) / gRadius2D;
          guessAmpl := min(maxAmpl, guessAmpl * 4 / separation);
        end
        else begin
          separation := 5;
          amplitude := maxAmpl;
          guessAmpl := maxAmpl;
        end;
      until (separation < 6) or (amplitude < 5.0E-3);
      if good and prevGood and ((key and kJumpedOdd) = (prevKey and
        kJumpedOdd)) then
        recDr(x1, x2, True);
      alpha := alpha + sign * amplitude;
      x1 := x2;
      prevKey := key;
      Application.ProcessMessages;
      if gStop then
        exit;
      separation := abs(alpha + sign * guessAmpl - halfAngle);
    until (separation > maxAngle);
  end;
end;


procedure DrawGeodesicBall(x1, vel: Vector3);
var
  u: Vector3;
  det, phi, topT0: Double;
begin
  case gType of
    Parametric, Tube:
      det := ComputeFrameParamSurface(vel, u, phi, topT0, x1);
    Implicit:
      det := ComputeFrameImplicitSurface(vel, u, phi, topT0, x1);
    PseudoRiemannian:
      det := ComputeFramePseudoRiemPlane(vel, u, phi, topT0, x1);
  else
    ;
  end;
  if det < 0 then begin
    DrawGeodesicQuadrant(x1, vel, u, topT0, -phi, phi);
    DrawGeodesicQuadrant(x1, vel, u, topT0, phi, pi - phi);
    DrawGeodesicQuadrant(x1, vel, u, topT0, pi - phi, pi + phi);
    DrawGeodesicQuadrant(x1, vel, u, topT0, pi + phi, 2 * pi - phi);
  end
  else if det > 0 then
    DrawGeodesicQuadrant(x1, vel, u, topT0, 0, 2 * pi + maxAmpl);
end;

function FindGeodesic(x1, x2, u1, u2: Vector3; alpha, beta, dAlpha, dBeta:
  Double): Boolean;
var
  angle, angle1, angle2, d, d1, d2: Double;
  co, si: Extended;
  k, key, counter: Integer;
  v: Vector3;
  good: Boolean;
begin
  result := False;
  angle1 := alpha;
  angle2 := beta;
  d1 := dAlpha;
  d2 := dBeta;
  good := False;
  counter := 0;
  repeat
    inc(counter);
    angle := (angle1 * d2 - angle2 * d1) / (d2 - d1);
    sincos(angle, si, co);
    for k := 1 to 3 do
      v[k] := u1[k] * co + u2[k] * si;
    key := theRunKut.GeodTowardsPoint(x1, x2, v);
    if (key and kMinimumReached) = 0 then break;
    d := theRunKut.distToLeft;
    if (d * d1 > 0) then begin
      angle1 := angle; d1 := d
    end
    else begin
      angle2 := angle; d2 := d
    end;
    good := abs(d) < theRunKut.radius;
  until good or (counter > 10);
  if good then begin
    theRunKut.topT := theRunKut.t;
    theRunKut.ComputeIntegralCurve(x1, v, 0);
    result := True;
  end;
end;

procedure DrawGeodesicBetweenPoints(x1, x2: Vector3);
const
  rayNumber = 61;
var
  p: Vector2;
  u1, u2, vel: Vector3;
  phi, E, F, G, aux, minDist, dist0, dist2, dist1: Double;
  si, co: Extended;
  i, k, key: Integer;
  good0, good1, good2: Boolean;
begin
  if (gType = Parametric) or (gType = Tube) then begin
    p[1] := x1[1]; p[2] := x1[2];
    ComputeEFG(p, E, F, G);
  end
  else if gType = PseudoRiemannian then begin
    p[1] := x1[1]; p[2] := x1[2];
    E := gTheF1.V(p);
    F := gTheF2.V(p);
    G := gTheF3.V(p);
  end;
  aux := 1;
  Randomize;
  u1[1] := Random;
  u1[2] := 1;
  u1[3] := 0;
  case gType of
    Parametric, Tube: begin
        ComputeFrameParamSurface(u1, u2, phi, aux, x1);
        theRunKut.SetEnvironment(2, 2, kTol, GeodAccelParametricSurf,
          CheckRangeParametric,
          RecordAndDrawSegmentParamSurface23D, nil, nil);
      end;
    Implicit: begin
        ComputeFrameImplicitSurface(u1, u2, phi, aux, x1);
        theRunKut.SetEnvironment(3, 2, kTol, GeodAccelImplicitSurf,
          CheckRangeImplicit,
          RecordAndDrawSegmentImplicitSurface, nil, AdjustPointImplicitSurf2);
      end;
    PseudoRiemannian: begin
        ComputeFramePseudoRiemPlane(u1, u2, phi, aux, x1);
        theRunKut.SetEnvironment(2, 2, kTol, GeodAccelPseudoRiemPlane,
          CheckRangePseudoRiem,
          RecordAndDrawSegmentPseudoRiemPlane, nil, nil);
      end;
  else
    ;
  end;
  minDist := EuclDistance3(x1, x2);
  dist0 := 2 * minDist;
  for i := 0 to rayNumber do begin
    phi := 2 * i * Pi / rayNumber;
    sincos(phi, si, co);
    for k := 1 to 3 do
      vel[k] := u1[k] * co + u2[k] * si;
    key := theRunKut.GeodTowardsPoint(x1, x2, vel);
    if i = 0 then begin
      good0 := (key and kMinimumReached) <> 0;
      if good0 then begin
        dist0 := theRunKut.distToLeft;
        dist2 := dist0;
        good2 := good0;
        if abs(dist2) < abs(minDist) then
          minDist := dist2;
      end;
    end
    else begin
      good1 := good2; dist1 := dist2;
      good2 := (key and kMinimumReached) <> 0;
      if good2 then begin
        dist2 := theRunKut.distToLeft;
        if abs(dist2) < abs(minDist) then
          minDist := dist2;
        if good1 and good2 and (dist1 * dist2 < 0) and
          FindGeodesic(x1, x2, u1, u2, phi - 2 * Pi / rayNumber, phi, dist1,
          dist2) then
          break;
      end;
      if (i = rayNumber) and good2 and good0 and (dist2 * dist0 < 0) and
        FindGeodesic(x1, x2, u1, u2, 2 * Pi / rayNumber, 0, dist2, dist0) then
        break;
    end
  end;
end;


procedure DrawGeodesic(x, vel: Vector3; toRight: integer);
begin
  case gType of
    Parametric, Tube:
      theRunKut.SetEnvironment(2, 2, kTol,
        GeodAccelParametricSurf, CheckRangeParametric,
        RecordAndDrawSegmentParamSurface23D,
        CheckCameBackParametricSurface, nil);
    Implicit:
      theRunKut.SetEnvironment(3, 2, kTol,
        GeodAccelImplicitSurf, CheckRangeImplicit,
        RecordAndDrawSegmentImplicitSurface,
        CheckCameBackImplicitSurface, AdjustPointImplicitSurf2);
    PseudoRiemannian:
      theRunKut.SetEnvironment(2, 2, kTol,
        GeodAccelPseudoRiemPlane, CheckRangePseudoRiem,
        RecordAndDrawSegmentPseudoRiemPlane, CameBackPseudoRiemPlane, nil);
  end;
  theRunKut.ComputeIntegralCurve(x, vel, toRight);
end;

procedure DrawSecondOrderCurve(toRight: Integer);
var
  x1, x2: Vector3;
  i, turn: Integer;
begin
  if gGeomOption = LinesOfConstGeodCurv then
    turn := toRight
  else
    turn := 0;
  if gGeomOption = GeodBetweenPts then begin
    if gType <> Implicit then begin
      for i := 1 to 2 do begin
        x1[i] := gFirst2DPos[i];
        x2[i] := gCur2DPos[i];
      end;
      x1[3] := 0; x2[3] := 0;
    end
    else begin
      x1 := gFirst3DPos;
      x2 := gCur3DPos;
    end;
  end
  else begin
    if gType <> Implicit then begin
      for i := 1 to 2 do
        x1[i] := gFirst2DPos[i];
      x1[3] := 0;
    end
    else
      x1 := gFirst3DPos;
  end;
  case gGeomOption of
    Geodesics, LinesOfConstGeodCurv:
      DrawGeodesic(x1, gVel, turn);
    ExponentialMap:
      DrawGeodesic(x1, gVel, 1);
    GeodBetweenPts:
      DrawGeodesicBetweenPoints(x1, x2);
    GeodesicBall:
      DrawGeodesicBall(x1, gVel);
  else
    ;
  end;
end;


end.
